package com.baidu.cloud.videoplayer.demo.feed;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.baidu.cloud.media.player.feature.FeatureException;
import com.baidu.cloud.media.player.feature.FeatureManager;
import com.baidu.cloud.videocache.ProxyCacheManager;
import com.baidu.cloud.videoplayer.demo.BaseApp;
import com.baidu.cloud.videoplayer.demo.R;
import com.baidu.cloud.videoplayer.demo.info.SharedPrefsStore;
import com.baidu.cloud.videoplayer.demo.info.VideoInfo;
import com.baidu.cloud.videoplayer.widget.BDCloudVideoView;

import android.content.Context;
import android.content.Intent;
import android.os.Bundle;
import android.view.View;
import android.view.ViewParent;
import android.widget.FrameLayout;
import android.widget.Toast;
import androidx.annotation.Nullable;
import androidx.appcompat.app.AppCompatActivity;
import androidx.viewpager.widget.ViewPager;

public class FeedActivity extends AppCompatActivity {
    private static final String TAG = "FeedActivity";

    private int mCurPos;
    private List<VideoInfo> mVideoList = new ArrayList<>();
    private FeedAdapter mFeedAdapter;
    private VerticalViewPager mViewPager;

    private ProxyCacheManager mCacheManager;

    private BDCloudVideoView mVideoView;

    private static final String KEY_INDEX = "index";

    public static void start(Context context, int index) {
        Intent i = new Intent(context, FeedActivity.class);
        i.putExtra(KEY_INDEX, index);
        context.startActivity(i);
    }

    @Override
    public void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_feed);
        initView();
    }

    private void initView() {
        initViewPager();
        initVideoView();
        mCacheManager = ProxyCacheManager.getInstance();
        addData(null);
        Intent extras = getIntent();
        final int index = extras.getIntExtra(KEY_INDEX, 0);
        mViewPager.setCurrentItem(index);

        mViewPager.post(new Runnable() {
            @Override
            public void run() {
                startPlay(index);
            }
        });
    }

    private void initVideoView() {
        BDCloudVideoView.setAppId(BaseApp.APP_ID);
        mVideoView = new BDCloudVideoView(this);
        mVideoView.setLooping(true);
    }

    private void initViewPager() {
        mViewPager = findViewById(R.id.vert_view_pager);
        mViewPager.setOffscreenPageLimit(4);
        mFeedAdapter = new FeedAdapter(mVideoList);
        mViewPager.setAdapter(mFeedAdapter);
        mViewPager.setOverScrollMode(View.OVER_SCROLL_NEVER);
        mViewPager.setOnPageChangeListener(new ViewPager.SimpleOnPageChangeListener() {

            private int mCurItem;

            /**
             * VerticalViewPager是否反向滑动
             */
            private boolean mIsReverseScroll;

            @Override
            public void onPageScrolled(int position, float positionOffset, int positionOffsetPixels) {
                super.onPageScrolled(position, positionOffset, positionOffsetPixels);
                if (position == mCurItem) {
                    return;
                }
                mIsReverseScroll = position < mCurItem;
            }

            @Override
            public void onPageSelected(int position) {
                super.onPageSelected(position);
                if (position == mCurPos) {
                    return;
                }
                startPlay(position);
            }

            @Override
            public void onPageScrollStateChanged(int state) {
                super.onPageScrollStateChanged(state);
                if (state == VerticalViewPager.SCROLL_STATE_DRAGGING) {
                    mCurItem = mViewPager.getCurrentItem();
                }

                if (state == VerticalViewPager.SCROLL_STATE_IDLE) {
                    mCacheManager.resumePreloadTask(mCurPos, mIsReverseScroll);
                } else {
                    mCacheManager.pausePreloadTask(mCurPos, mIsReverseScroll);
                }
            }
        });
    }

    private void startPlay(int position) {
        int count = mViewPager.getChildCount();
        for (int i = 0; i < count; i ++) {
            View itemView = mViewPager.getChildAt(i);
            FeedAdapter.ViewHolder viewHolder = (FeedAdapter.ViewHolder) itemView.getTag();
            if (viewHolder.mPosition == position) {
                mVideoView.release();
                removeViewFormParent(mVideoView);

                VideoInfo videoInfo = mVideoList.get(position);
                prepareVideoViewForVR(videoInfo);
                mVideoView.reSetRender();

                String playUrl = mCacheManager.getPlayableUrl(this, videoInfo.getUrl());
                mVideoView.setVideoPath(playUrl);

                viewHolder.mViewHolderRoot.addView(mVideoView, 0);
                viewHolder.setVideoView(mVideoView);

                mVideoView.start();
                mCurPos = position;
                break;
            }
        }
    }

    public void addData(View view) {
        mVideoList.addAll(SharedPrefsStore.getAllFeedVideoFromSP(this));
        mFeedAdapter.notifyDataSetChanged();
    }

    protected void onResume() {
        super.onResume();
        if (mVideoView != null) {
            mVideoView.start();
        }
    }


    @Override
    protected void onPause() {
        super.onPause();
        if (mVideoView != null) {
            mVideoView.pause();
        }
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();
        mCacheManager.removeAllPreloadTask();
        // 也可以不清除，业务按需处理
        try {
            ProxyCacheManager.getInstance().cleanVideoCacheDir(this);
        } catch (IOException e) {
            e.printStackTrace();
        }
        if (mVideoView != null) {
            mVideoView.release();
        }
    }

    private static void removeViewFormParent(View v) {
        if (v == null) {
            return;
        }
        ViewParent parent = v.getParent();
        if (parent instanceof FrameLayout) {
            ((FrameLayout) parent).removeView(v);
        }
    }

    private void prepareVideoViewForVR(VideoInfo info) {
        if (info.isVrEnable()) {
            try {
                if (FeatureManager.getInstance().checkVrFeatureValid(this, BaseApp.APP_ID)) {
                    BDCloudVideoView.setVrEnable(info.isVrEnable());
                }
            } catch (FeatureException e) {
                if (e.getFeatureErrCode() == FeatureException.FEATURE_ERR_LICENSE_INVALID) {
                    Toast.makeText(this, "当前License不支持VR播放", Toast.LENGTH_SHORT).show();
                } else {
                    Toast.makeText(this, "当前SDK不支持VR播放功能", Toast.LENGTH_SHORT).show();
                }
                BDCloudVideoView.setVrEnable(false);
            }
        } else {
            BDCloudVideoView.setVrEnable(false);
        }
    }
}