package com.baidu.cloud.videoplayer.demo.feed;

import android.content.Context;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.FrameLayout;
import android.widget.ImageButton;
import android.widget.ImageView;
import android.widget.TextView;

import androidx.annotation.NonNull;
import androidx.viewpager.widget.PagerAdapter;

import com.baidu.cloud.videocache.ProxyCacheManager;
import com.baidu.cloud.videoplayer.demo.R;
import com.baidu.cloud.videoplayer.demo.info.VideoInfo;
import com.baidu.cloud.videoplayer.widget.BDCloudVideoView;
import com.bumptech.glide.Glide;

import java.util.ArrayList;
import java.util.List;

public class FeedAdapter extends PagerAdapter {
    private static final String TAG = "FeedAdapter";

    /**
     * View缓存池，从ViewPager中移除的item将会存到这里面，用来复用
     */
    private List<View> mViewPool = new ArrayList<>();
    private List<VideoInfo> mVideoInfos;

    public FeedAdapter(List<VideoInfo> videoBeans) {
        this.mVideoInfos = videoBeans;
    }

    @Override
    public int getCount() {
        return mVideoInfos == null ? 0 : mVideoInfos.size();
    }

    @Override
    public boolean isViewFromObject(@NonNull View view, @NonNull Object o) {
        return view == o;
    }

    @NonNull
    @Override
    public Object instantiateItem(@NonNull ViewGroup container, int position) {
        final Context context = container.getContext();
        View view = null;
        if (mViewPool.size() > 0) { // 取第一个进行复用
            view = mViewPool.get(0);
            mViewPool.remove(0);
        }

        ViewHolder viewHolder;
        if (view == null) {
            view = LayoutInflater.from(context).inflate(R.layout.item_of_feed, container, false);
            viewHolder = new ViewHolder(view);
        } else {
            viewHolder = (ViewHolder) view.getTag();
        }

        VideoInfo info = mVideoInfos.get(position);
        // 开始预加载
        ProxyCacheManager.getInstance().addPreloadTask(context, info.getUrl(), position);
        Glide.with(context)
                .load(info.getImageUrl())
                .placeholder(android.R.color.black)
                .into(viewHolder.mCover);
        viewHolder.mTitle.setText(info.getTitle());
        viewHolder.mPosition = position;
        container.addView(view);
        return view;
    }

    @Override
    public void destroyItem(@NonNull ViewGroup container, int position, @NonNull Object object) {
        View itemView = (View) object;
        container.removeView(itemView);
        VideoInfo info = mVideoInfos.get(position);
        // 取消预加载
        ProxyCacheManager.getInstance().removePreloadTask(info.getUrl());
        // 保存起来用来复用
        mViewPool.add(itemView);
    }

    /**
     * 借鉴ListView item复用方法
     */
    public static class ViewHolder {

        public int mPosition;
        public TextView mTitle;
        public ImageView mCover;
        public ImageButton mPlayBtn;
        public FrameLayout mViewHolderRoot;
        private BDCloudVideoView mVideoView;

        ViewHolder(View itemView) {
            mTitle = itemView.findViewById(R.id.tv_title);
            mCover = itemView.findViewById(R.id.iv_cover);
            mPlayBtn = itemView.findViewById(R.id.play_btn);
            mViewHolderRoot = itemView.findViewById(R.id.view_holder_root);

            itemView.setTag(this);
        }

        public void setVideoView(BDCloudVideoView videoView) {
            mVideoView = videoView;
            mVideoView.setOnPlayerStateListener(new BDCloudVideoView.OnPlayerStateListener() {
                @Override
                public void onPlayerStateChanged(BDCloudVideoView.PlayerState nowState) {
                    switch (nowState) {
                        case STATE_IDLE:
                            mCover.setVisibility(View.VISIBLE);
                            break;
                        case STATE_PLAYING:
                            mCover.setVisibility(View.GONE);
                            mPlayBtn.setVisibility(View.GONE);
                            break;
                        case STATE_PAUSED:
                            mCover.setVisibility(View.GONE);
                            mPlayBtn.setVisibility(View.VISIBLE);
                            break;
                        default:
                            break;
                    }
                }
            });
            mViewHolderRoot.setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View v) {
                    togglePlayPause();
                }
            });
            mPlayBtn.setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View v) {
                    togglePlayPause();
                }
            });
        }

        private void togglePlayPause() {
            if (mVideoView != null) {
                if (mVideoView.isPlaying()) {
                    mVideoView.pause();
                } else  {
                    mVideoView.start();
                }
            }
        }
    }
}
