package com.baidu.cloud.videoplayer.widget;

import android.graphics.Bitmap;
import android.graphics.SurfaceTexture;
import android.opengl.GLES11Ext;
import android.opengl.GLES30;
import android.opengl.Matrix;
import android.util.Log;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;

public class GLTextureRender {
    private static final String TAG = "GLTextureRender";

    private static final int FLOAT_SIZE_BYTES = 4;
    private static final int TRIANGLE_VERTICES_DATA_STRIDE_BYTES = 5 * FLOAT_SIZE_BYTES;
    private static final int TRIANGLE_VERTICES_DATA_POS_OFFSET = 0;
    private static final int TRIANGLE_VERTICES_DATA_UV_OFFSET = 3;
    private final float[] mTriangleVerticesData = {
            // X, Y, Z, U, V
            -1.0f, -1.0f, 0, 0.f, 0.f,
            1.0f, -1.0f, 0, 1.f, 0.f,
            -1.0f,  1.0f, 0, 0.f, 1.f,
            1.0f,  1.0f, 0, 1.f, 1.f,
    };

    private final float[] mVerticesData = {
            // X, Y, Z
            -1.0f, -1.0f, 0,
            1.0f, -1.0f, 0,
            -1.0f, 1.0f, 0,
            1.0f, 1.0f, 0
    };

    private final float[] mTextureData = {
            // U, V
            0.f, 0.f,
            1.f, 0.f,
            0.f, 1.f,
            1.f, 1.f,
    };

    private FloatBuffer mTriangleVertices;
    private FloatBuffer mVerticesBuffer;
    private FloatBuffer mTexturesBuffer;

    private static final String VERTEX_SHADER =
            "#version 300 es\n" +
                    "uniform mat4 uMVPMatrix;\n" +
                    "uniform mat4 uSTMatrix;\n" +
                    "layout (location = 0) in vec4 aPosition;\n" +
                    "layout (location = 1) in vec4 aTextureCoord;\n" +
                    "out vec2 vTextureCoord;\n" +
                    "void main() {\n" +
                    "   gl_Position = uMVPMatrix * aPosition;\n" +
                    "   vTextureCoord = (uSTMatrix * aTextureCoord).xy;\n" +
                    "}\n";

    private static final String FRAGMENT_SHADER =
            "#version 300 es\n" +
                    "#extension GL_OES_EGL_image_external_essl3 : require\n" +
                    "precision mediump float;\n" +
                    "in vec2 vTextureCoord;\n" +
                    "uniform samplerExternalOES sTexture;\n" +
                    "out vec4 color;\n" +
                    "void main() {\n" +
                    "  color = texture(sTexture, vTextureCoord);\n" +
                    "}\n";

    private static final String FRAGMENT_SHADER_TEX =
            "#version 300 es\n" +
                    "#extension GL_OES_EGL_image_external_essl3 : require\n" +
                    "precision mediump float;\n" +
                    "in vec2 vTextureCoord;\n" +
                    "uniform sampler2D sTexture;\n" +
                    "out vec4 color;\n" +
                    "void main() {\n" +
                    "  color = texture(sTexture, vTextureCoord);\n" +
                    "}\n";

    private static final String VERTEX_SHADER_MIX =
            "#version 300 es\n" +
                    "uniform mat4 uMVPMatrix;\n" +
                    "uniform mat4 uSTMatrix;\n" +
                    "uniform mat4 uSTMatrix2;\n" +
                    "layout (location = 0) in vec4 aPosition;\n" +
                    "layout (location = 1) in vec4 aTextureCoord;\n" +
                    "layout (location = 2) in vec4 aTextureCoord2;\n" +
                    "out vec2 vTextureCoord;\n" +
                    "out vec2 vTextureCoord2;\n" +
                    "void main() {\n" +
                    "   gl_Position = uMVPMatrix * aPosition;\n" +
                    "   vTextureCoord = (uSTMatrix * aTextureCoord).xy;\n" +
                    "   vTextureCoord2 = (uSTMatrix2 * aTextureCoord2).xy;\n" +
                    "}\n";

    private static final String FRAGMENT_SHADER_MIX =
            "#version 300 es\n" +
                    "#extension GL_OES_EGL_image_external_essl3 : require\n" +
                    "precision mediump float;\n" +
                    "in vec2 vTextureCoord;\n" +
                    "in vec2 vTextureCoord2;\n" +
                    "out vec4 color;\n" +
                    "uniform samplerExternalOES input_texture;\n" +
                    "uniform sampler2D colorkey_texture;\n" +
                    "void main() {\n" +
                    "vec4 bg = texture(input_texture, vTextureCoord);\n" +
                    "vec4 fg = texture(colorkey_texture, vTextureCoord2);\n" +
                    "color = vec4(mix(bg.rgb, fg.rgb, fg.a), 1.0);\n" +
                    "}\n";

    private float[] mMVPMatrix = new float[16];
    private float[] mSTMatrix = new float[16];
    private float[] mSTMatrix2 = new float[16];
    private float[] mSTMatrixMix = new float[16];
    private float[] mSTMatrixMix2 = new float[16];

    private int mProgram;
    private int mProgramTex2Screen;
    private int mProgramMix;
    private int mTextureID = -12345;
    private int muMVPMatrixHandle;
    private int muSTMatrixHandle;
    private int maPositionHandle;
    private int maTextureHandle;

    private int mVaoOES;
    private int mVao2D;
    private int mVaoMix;
    private boolean useVAO = true;

    private int muMVPMatrixHandleTex2Screen;
    private int muSTMatrixHandleTex2Screen;
    private int maPositionHandleTex2Screen;
    private int maTextureHandleTex2Screen;

    private int muMVPMatrixHandleMix;
    private int muSTMatrixHandleMix;
    private int muSTMatrixHandleMix2;
    private int maPositionHandleMix;
    private int maTextureHandleMix;
    private int maTextureHandleMix2;

    public GLTextureRender() {
        mTriangleVertices = ByteBuffer.allocateDirect(
                        mTriangleVerticesData.length * FLOAT_SIZE_BYTES)
                .order(ByteOrder.nativeOrder()).asFloatBuffer();
        mTriangleVertices.put(mTriangleVerticesData).position(0);

        mVerticesBuffer = ByteBuffer.allocateDirect(
                        mVerticesData.length * FLOAT_SIZE_BYTES)
                .order(ByteOrder.nativeOrder()).asFloatBuffer();
        mVerticesBuffer.put(mVerticesData).position(0);

        mTexturesBuffer = ByteBuffer.allocateDirect(
                        mTextureData.length * FLOAT_SIZE_BYTES)
                .order(ByteOrder.nativeOrder()).asFloatBuffer();
        mTexturesBuffer.put(mTextureData).position(0);

        Matrix.setIdentityM(mSTMatrix, 0);
    }

    public int getTextureId() {
        return mTextureID;
    }

    public void drawFrame(SurfaceTexture st, int targetfbo) {
        drawFrame(st, mTextureID, targetfbo);
    }

    public void drawFrame(SurfaceTexture st, int textureId, int targetfbo) {
        checkGlError("onDrawFrame startRenderToSurface");
        st.getTransformMatrix(mSTMatrix);

        GLES30.glBindFramebuffer(GLES30.GL_FRAMEBUFFER, targetfbo);
        GLES30.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        GLES30.glClear(GLES30.GL_DEPTH_BUFFER_BIT | GLES30.GL_COLOR_BUFFER_BIT);

        GLES30.glUseProgram(mProgram);
        checkGlError("glUseProgram");

        GLES30.glActiveTexture(GLES30.GL_TEXTURE0);
        GLES30.glBindTexture(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, textureId);

        if (!useVAO) {
            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_POS_OFFSET);
            GLES30.glVertexAttribPointer(maPositionHandle, 3, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maPosition");
            GLES30.glEnableVertexAttribArray(maPositionHandle);
            checkGlError("glEnableVertexAttribArray maPositionHandle");

            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_UV_OFFSET);
            GLES30.glVertexAttribPointer(maTextureHandle, 2, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maTextureHandle");
            GLES30.glEnableVertexAttribArray(maTextureHandle);
            checkGlError("glEnableVertexAttribArray maTextureHandle");
        }

        Matrix.setIdentityM(mMVPMatrix, 0);

        GLES30.glUniformMatrix4fv(muMVPMatrixHandle, 1, false, mMVPMatrix, 0);
        GLES30.glUniformMatrix4fv(muSTMatrixHandle, 1, false, mSTMatrix, 0);

        if (useVAO) {
            GLES30.glBindVertexArray(mVaoOES);
        }
        GLES30.glDrawArrays(GLES30.GL_TRIANGLE_STRIP, 0, 4);
        checkGlError("glDrawArrays");
        if (useVAO) {
            GLES30.glBindVertexArray(0);
        }
        GLES30.glFinish();
    }

    public void drawFrame(int texId, int targetfbo) {
        checkGlError("onDrawFrame startRenderToSurface");
        Matrix.setIdentityM(mSTMatrix2, 0);

        GLES30.glBindFramebuffer(GLES30.GL_FRAMEBUFFER, targetfbo);
        GLES30.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        GLES30.glClear(GLES30.GL_DEPTH_BUFFER_BIT | GLES30.GL_COLOR_BUFFER_BIT);

        GLES30.glUseProgram(mProgramTex2Screen);
        checkGlError("glUseProgram");

        GLES30.glActiveTexture(GLES30.GL_TEXTURE0);
        GLES30.glBindTexture(GLES30.GL_TEXTURE_2D, texId);

        if (!useVAO) {
            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_POS_OFFSET);
            GLES30.glVertexAttribPointer(maPositionHandleTex2Screen, 3, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maPosition");
            GLES30.glEnableVertexAttribArray(maPositionHandleTex2Screen);
            checkGlError("glEnableVertexAttribArray maPositionHandle");

            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_UV_OFFSET);
            GLES30.glVertexAttribPointer(maTextureHandleTex2Screen, 2, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maTextureHandle");
            GLES30.glEnableVertexAttribArray(maTextureHandleTex2Screen);
            checkGlError("glEnableVertexAttribArray maTextureHandle");
        }

        Matrix.setIdentityM(mMVPMatrix, 0);

        GLES30.glUniformMatrix4fv(muMVPMatrixHandleTex2Screen, 1, false, mMVPMatrix, 0);
        GLES30.glUniformMatrix4fv(muSTMatrixHandleTex2Screen, 1, false, mSTMatrix2, 0);

        if (useVAO) {
            GLES30.glBindVertexArray(mVao2D);
        }
        GLES30.glDrawArrays(GLES30.GL_TRIANGLE_STRIP, 0, 4);
        checkGlError("glDrawArrays");
        if (useVAO) {
            GLES30.glBindVertexArray(0);
        }
        GLES30.glFinish();
    }

    public void mixSurfaceAndTex(SurfaceTexture st, int texid, int targetfbo) {
        checkGlError("onDrawFrame startMixSurfaceAndTex");
        st.getTransformMatrix(mSTMatrixMix);
        Matrix.setIdentityM(mSTMatrixMix2, 0);

        GLES30.glBindFramebuffer(GLES30.GL_FRAMEBUFFER, targetfbo);
        GLES30.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        GLES30.glClear(GLES30.GL_DEPTH_BUFFER_BIT | GLES30.GL_COLOR_BUFFER_BIT);

        GLES30.glUseProgram(mProgramMix);
        checkGlError("glUseProgram");

        GLES30.glActiveTexture(GLES30.GL_TEXTURE0);
        GLES30.glBindTexture(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, mTextureID);
        int location = GLES30.glGetUniformLocation(mProgramMix, "input_texture");
        GLES30.glUniform1i(location, 0);

        GLES30.glActiveTexture(GLES30.GL_TEXTURE1);
        GLES30.glBindTexture(GLES30.GL_TEXTURE_2D, texid);
        location = GLES30.glGetUniformLocation(mProgramMix, "colorkey_texture");
        GLES30.glUniform1i(location, 1);

        if (!useVAO) {
            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_POS_OFFSET);
            GLES30.glVertexAttribPointer(maPositionHandleMix, 3, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maPosition");
            GLES30.glEnableVertexAttribArray(maPositionHandleMix);
            checkGlError("glEnableVertexAttribArray maPositionHandle");

            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_UV_OFFSET);
            GLES30.glVertexAttribPointer(maTextureHandleMix, 2, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maTextureHandle");
            GLES30.glEnableVertexAttribArray(maTextureHandleMix);
            checkGlError("glEnableVertexAttribArray maTextureHandle");

            mTriangleVertices.position(TRIANGLE_VERTICES_DATA_UV_OFFSET);
            GLES30.glVertexAttribPointer(maTextureHandleMix2, 2, GLES30.GL_FLOAT, false,
                    TRIANGLE_VERTICES_DATA_STRIDE_BYTES, mTriangleVertices);
            checkGlError("glVertexAttribPointer maTextureHandle");
            GLES30.glEnableVertexAttribArray(maTextureHandleMix2);
            checkGlError("glEnableVertexAttribArray maTextureHandle");
        }

        Matrix.setIdentityM(mMVPMatrix, 0);

        GLES30.glUniformMatrix4fv(muMVPMatrixHandleMix, 1, false, mMVPMatrix, 0);
        GLES30.glUniformMatrix4fv(muSTMatrixHandleMix, 1, false, mSTMatrixMix, 0);
        GLES30.glUniformMatrix4fv(muSTMatrixHandleMix2, 1, false, mSTMatrixMix2, 0);

        if (useVAO) {
            GLES30.glBindVertexArray(mVaoMix);
        }
        GLES30.glDrawArrays(GLES30.GL_TRIANGLE_STRIP, 0, 4);
        checkGlError("glDrawArrays");
        if (useVAO) {
            GLES30.glBindVertexArray(0);
        }
        GLES30.glFinish();
    }

    /**
     * Initializes GL state.  Call this after the EGL surface has been created and made current.
     */
    public void surfaceCreated() {
        /**
         * render oes
         */
        mProgram = createProgram(VERTEX_SHADER, FRAGMENT_SHADER);
        if (mProgram == 0) {
            throw new RuntimeException("failed creating program");
        }
        maPositionHandle = GLES30.glGetAttribLocation(mProgram, "aPosition");
        checkGlError("glGetAttribLocation aPosition");
        if (maPositionHandle == -1) {
            throw new RuntimeException("Could not get attrib location for aPosition");
        }
        maTextureHandle = GLES30.glGetAttribLocation(mProgram, "aTextureCoord");
        checkGlError("glGetAttribLocation aTextureCoord");
        if (maTextureHandle == -1) {
            throw new RuntimeException("Could not get attrib location for aTextureCoord");
        }

        muMVPMatrixHandle = GLES30.glGetUniformLocation(mProgram, "uMVPMatrix");
        checkGlError("glGetUniformLocation uMVPMatrix");
        if (muMVPMatrixHandle == -1) {
            throw new RuntimeException("Could not get attrib location for uMVPMatrix");
        }

        muSTMatrixHandle = GLES30.glGetUniformLocation(mProgram, "uSTMatrix");
        checkGlError("glGetUniformLocation uSTMatrix");
        if (muSTMatrixHandle == -1) {
            throw new RuntimeException("Could not get attrib location for uSTMatrix");
        }

        /**
         * render 2d
         */
        mProgramTex2Screen = createProgram(VERTEX_SHADER, FRAGMENT_SHADER_TEX);
        if (mProgramTex2Screen == 0) {
            throw new RuntimeException("failed creating program");
        }
        maPositionHandleTex2Screen = GLES30.glGetAttribLocation(mProgramTex2Screen, "aPosition");
        checkGlError("glGetAttribLocation aPosition");
        if (maPositionHandleTex2Screen == -1) {
            throw new RuntimeException("Could not get attrib location for aPosition");
        }
        maTextureHandleTex2Screen = GLES30.glGetAttribLocation(mProgramTex2Screen, "aTextureCoord");
        checkGlError("glGetAttribLocation aTextureCoord");
        if (maTextureHandleTex2Screen == -1) {
            throw new RuntimeException("Could not get attrib location for aTextureCoord");
        }

        muMVPMatrixHandleTex2Screen = GLES30.glGetUniformLocation(mProgramTex2Screen, "uMVPMatrix");
        checkGlError("glGetUniformLocation uMVPMatrix");
        if (muMVPMatrixHandleTex2Screen == -1) {
            throw new RuntimeException("Could not get attrib location for uMVPMatrix");
        }

        muSTMatrixHandleTex2Screen = GLES30.glGetUniformLocation(mProgramTex2Screen, "uSTMatrix");
        checkGlError("glGetUniformLocation uSTMatrix");
        if (muSTMatrixHandleTex2Screen == -1) {
            throw new RuntimeException("Could not get attrib location for uSTMatrix");
        }

        /**
         * mix oes and 2d
         */
        mProgramMix = createProgram(VERTEX_SHADER_MIX, FRAGMENT_SHADER_MIX);
        if (mProgramMix == 0) {
            throw new RuntimeException("failed creating program");
        }
        maPositionHandleMix = GLES30.glGetAttribLocation(mProgramMix, "aPosition");
        checkGlError("glGetAttribLocation aPosition");
        if (maPositionHandleMix == -1) {
            throw new RuntimeException("Could not get attrib location for aPosition");
        }
        maTextureHandleMix = GLES30.glGetAttribLocation(mProgramMix, "aTextureCoord");
        checkGlError("glGetAttribLocation aTextureCoord");
        if (maTextureHandleMix == -1) {
            throw new RuntimeException("Could not get attrib location for aTextureCoord");
        }
        maTextureHandleMix2 = GLES30.glGetAttribLocation(mProgramMix, "aTextureCoord2");
        checkGlError("glGetAttribLocation aTextureCoord2");
        if (maTextureHandleMix == -1) {
            throw new RuntimeException("Could not get attrib location for aTextureCoord2");
        }

        muMVPMatrixHandleMix = GLES30.glGetUniformLocation(mProgramMix, "uMVPMatrix");
        checkGlError("glGetUniformLocation uMVPMatrix");
        if (muMVPMatrixHandleMix == -1) {
            throw new RuntimeException("Could not get attrib location for uMVPMatrix");
        }

        muSTMatrixHandleMix = GLES30.glGetUniformLocation(mProgramMix, "uSTMatrix");
        checkGlError("glGetUniformLocation uSTMatrix");
        if (muSTMatrixHandleMix == -1) {
            throw new RuntimeException("Could not get attrib location for uSTMatrix");
        }

        muSTMatrixHandleMix2 = GLES30.glGetUniformLocation(mProgramMix, "uSTMatrix2");
        checkGlError("glGetUniformLocation uSTMatrix2");
        if (muSTMatrixHandleMix2 == -1) {
            throw new RuntimeException("Could not get attrib location for uSTMatrix2");
        }

        int[] textures = new int[1];
        GLES30.glGenTextures(1, textures, 0);

        mTextureID = textures[0];
        GLES30.glBindTexture(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, mTextureID);
        checkGlError("glBindTexture mTextureID");

        GLES30.glTexParameterf(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, GLES30.GL_TEXTURE_MIN_FILTER,
                GLES30.GL_LINEAR);
        GLES30.glTexParameterf(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, GLES30.GL_TEXTURE_MAG_FILTER,
                GLES30.GL_LINEAR);
        GLES30.glTexParameteri(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, GLES30.GL_TEXTURE_WRAP_S,
                GLES30.GL_CLAMP_TO_EDGE);
        GLES30.glTexParameteri(GLES11Ext.GL_TEXTURE_EXTERNAL_OES, GLES30.GL_TEXTURE_WRAP_T,
                GLES30.GL_CLAMP_TO_EDGE);
        checkGlError("glTexParameter");

        if (!useVAO) {
            return;
        }
        mVaoOES = createVao();
        GLES30.glBindVertexArray(mVaoOES);
        int vertexVbo = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, vertexVbo);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mVerticesData.length * FLOAT_SIZE_BYTES,
                mVerticesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maPositionHandle, 3, GLES30.GL_FLOAT, false,
                mVerticesData.length, 0);
        checkGlError("glVertexAttribPointer maPosition");
        GLES30.glEnableVertexAttribArray(maPositionHandle);
        checkGlError("glEnableVertexAttribArray maPositionHandle");

        int texVbo = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, texVbo);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mTextureData.length * FLOAT_SIZE_BYTES,
                mTexturesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maTextureHandle, 2, GLES30.GL_FLOAT, false,
                mTextureData.length, 0);
        checkGlError("glVertexAttribPointer maTextureHandle");
        GLES30.glEnableVertexAttribArray(maTextureHandle);
        checkGlError("glEnableVertexAttribArray maTextureHandle");
        GLES30.glBindVertexArray(0);

        mVao2D = createVao();
        GLES30.glBindVertexArray(mVao2D);
        int vertexVbo2 = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, vertexVbo2);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mVerticesData.length * FLOAT_SIZE_BYTES,
                mVerticesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maPositionHandleTex2Screen, 3, GLES30.GL_FLOAT, false,
                mVerticesData.length, 0);
        checkGlError("glVertexAttribPointer maPosition");
        GLES30.glEnableVertexAttribArray(maPositionHandleTex2Screen);
        checkGlError("glEnableVertexAttribArray maPositionHandle");

        int texVbo2 = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, texVbo2);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mTextureData.length * FLOAT_SIZE_BYTES,
                mTexturesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maTextureHandleTex2Screen, 2, GLES30.GL_FLOAT, false,
                mTextureData.length, 0);
        checkGlError("glVertexAttribPointer maTextureHandle");
        GLES30.glEnableVertexAttribArray(maTextureHandleTex2Screen);
        checkGlError("glEnableVertexAttribArray maTextureHandle");
        GLES30.glBindVertexArray(0);

        mVaoMix = createVao();
        GLES30.glBindVertexArray(mVaoMix);
        int vertexVboMix = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, vertexVboMix);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mVerticesData.length * FLOAT_SIZE_BYTES,
                mVerticesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maPositionHandleMix, 3, GLES30.GL_FLOAT, false,
                mVerticesData.length, 0);
        checkGlError("glVertexAttribPointer maPosition");
        GLES30.glEnableVertexAttribArray(maPositionHandleMix);
        checkGlError("glEnableVertexAttribArray maPositionHandle");
        GLES30.glBindVertexArray(0);

        GLES30.glBindVertexArray(mVaoMix);
        int texVboMix = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, texVboMix);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mTextureData.length * FLOAT_SIZE_BYTES,
                mTexturesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maTextureHandleMix, 2, GLES30.GL_FLOAT, false,
                mTextureData.length, 0);
        checkGlError("glVertexAttribPointer maTextureHandle");
        GLES30.glEnableVertexAttribArray(maTextureHandleMix);
        checkGlError("glEnableVertexAttribArray maTextureHandle");
        GLES30.glBindVertexArray(0);

        GLES30.glBindVertexArray(mVaoMix);
        int texVboMix2 = createVbo();
        GLES30.glBindBuffer(GLES30.GL_ARRAY_BUFFER, texVboMix2);
        GLES30.glBufferData(GLES30.GL_ARRAY_BUFFER, mTextureData.length * FLOAT_SIZE_BYTES,
                mTexturesBuffer, GLES30.GL_STATIC_DRAW);
        GLES30.glVertexAttribPointer(maTextureHandleMix2, 2, GLES30.GL_FLOAT, false,
                mTextureData.length, 0);
        checkGlError("glVertexAttribPointer maTextureHandle");
        GLES30.glEnableVertexAttribArray(maTextureHandleMix2);
        checkGlError("glEnableVertexAttribArray maTextureHandle");
        GLES30.glBindVertexArray(0);
    }

    public void surfaceChanged(final int width, final int height) {
        GLES30.glViewport(0, 0, width, height);
    }

    /**
     * Replaces the fragment shader.
     */
    public void changeFragmentShader(String fragmentShader) {
        GLES30.glDeleteProgram(mProgram);
        mProgram = createProgram(VERTEX_SHADER, fragmentShader);
        if (mProgram == 0) {
            throw new RuntimeException("failed creating program");
        }
    }

    private int loadShader(int shaderType, String source) {
        int shader = GLES30.glCreateShader(shaderType);
        checkGlError("glCreateShader type=" + shaderType);
        GLES30.glShaderSource(shader, source);
        GLES30.glCompileShader(shader);
        int[] compiled = new int[1];
        GLES30.glGetShaderiv(shader, GLES30.GL_COMPILE_STATUS, compiled, 0);
        if (compiled[0] == 0) {
            Log.e(TAG, "Could not compile shader " + shaderType + ":");
            Log.e(TAG, " " + GLES30.glGetShaderInfoLog(shader));
            GLES30.glDeleteShader(shader);
            shader = 0;
        }
        return shader;
    }

    private int createProgram(String vertexSource, String fragmentSource) {
        int vertexShader = loadShader(GLES30.GL_VERTEX_SHADER, vertexSource);
        if (vertexShader == 0) {
            return 0;
        }
        int pixelShader = loadShader(GLES30.GL_FRAGMENT_SHADER, fragmentSource);
        if (pixelShader == 0) {
            return 0;
        }

        int program = GLES30.glCreateProgram();
        checkGlError("glCreateProgram");
        if (program == 0) {
            Log.e(TAG, "Could not create program");
        }
        GLES30.glAttachShader(program, vertexShader);
        checkGlError("glAttachShader");
        GLES30.glAttachShader(program, pixelShader);
        checkGlError("glAttachShader");
        GLES30.glLinkProgram(program);
        int[] linkStatus = new int[1];
        GLES30.glGetProgramiv(program, GLES30.GL_LINK_STATUS, linkStatus, 0);
        if (linkStatus[0] != GLES30.GL_TRUE) {
            Log.e(TAG, "Could not link program: ");
            Log.e(TAG, GLES30.glGetProgramInfoLog(program));
            GLES30.glDeleteProgram(program);
            program = 0;
        }
        return program;
    }

    public static void checkGlError(String op) {
        int error;
        while ((error = GLES30.glGetError()) != GLES30.GL_NO_ERROR) {
            Log.e(TAG, op + ": glError " + error);
            throw new RuntimeException(op + ": glError " + error);
        }
    }

    public static void destroyTextureObject(int textureId) {
        if (textureId == -1) {
            return;
        }
        GLES30.glDeleteTextures(1, new int[] {textureId}, 0);
        checkGlError("glDeleteTextures");
    }

    public static int createTextureObject(int textureTarget, int width, int height) {
        int texId = createTextureObject(textureTarget);
        GLES30.glTexImage2D(textureTarget, 0, GLES30.GL_RGBA, width, height, 0,
                GLES30.GL_RGBA, GLES30.GL_UNSIGNED_BYTE, null);

        return texId;
    }

    public static int createTextureObject(int textureTarget) {
        int[] textures = new int[1];
        GLES30.glGenTextures(1, textures, 0);
        checkGlError("glGenTextures");

        int texId = textures[0];
        GLES30.glBindTexture(textureTarget, texId);
        checkGlError("glBindTexture " + texId);

        GLES30.glTexParameterf(textureTarget, GLES30.GL_TEXTURE_MIN_FILTER, GLES30.GL_LINEAR);
        GLES30.glTexParameterf(textureTarget, GLES30.GL_TEXTURE_MAG_FILTER, GLES30.GL_LINEAR);
        GLES30.glTexParameteri(textureTarget, GLES30.GL_TEXTURE_WRAP_S, GLES30.GL_CLAMP_TO_EDGE);
        GLES30.glTexParameteri(textureTarget, GLES30.GL_TEXTURE_WRAP_T, GLES30.GL_CLAMP_TO_EDGE);
        checkGlError("glTexParameter");

        return texId;
    }

    public static int createFrameBufferObject() {
        int[] tmpId = new int[1];
        GLES30.glGenFramebuffers(1, tmpId, 0);
        return tmpId[0];
    }

    public static int createVao() {
        int[] tmpId = new int[1];
        GLES30.glGenVertexArrays(1, tmpId, 0);
        return tmpId[0];
    }

    public static int createVbo() {
        int[] tmpId = new int[1];
        GLES30.glGenBuffers(1, tmpId, 0);
        return tmpId[0];
    }

    public static void bindTextureToFBO(int textureId, int textureTarget, int fboId) {
        GLES30.glBindFramebuffer(GLES30.GL_FRAMEBUFFER, fboId);
        GLES30.glClearColor(0,  0,  0,  0);
        GLES30.glBindTexture(textureTarget, textureId);
        GLES30.glFramebufferTexture2D(GLES30.GL_FRAMEBUFFER, GLES30.GL_COLOR_ATTACHMENT0
                , textureTarget, textureId, 0);
    }

    public Bitmap getBitmap(int width, int height) {
        // glReadPixels gives us a ByteBuffer filled with what is essentially big-endian RGBA
        // data (i.e. a byte of red, followed by a byte of green...).  We need an int[] filled
        // with native-order ARGB data to feed to Bitmap.
        //
        // If we implement this as a series of buf.get() calls, we can spend 2.5 seconds just
        // copying data around for a 720p frame.  It's better to do a bulk get() and then
        // rearrange the data in memory.  (For comparison, the PNG compress takes about 500ms
        // for a trivial frame.)
        //
        // So... we set the ByteBuffer to little-endian, which should turn the bulk IntBuffer
        // get() into a straight memcpy on most Android devices.  Our ints will hold ABGR data.
        // Swapping B and R gives us ARGB.  We need about 30ms for the bulk get(), and another
        // 270ms for the color swap.
        //
        // Making this even more interesting is the upside-down nature of GL, which means we
        // may want to flip the image vertically here.

        ByteBuffer buf = ByteBuffer.allocateDirect(width * height * 4);
        buf.order(ByteOrder.LITTLE_ENDIAN);
        GLES30.glReadPixels(0, 0, width, height, GLES30.GL_RGBA, GLES30.GL_UNSIGNED_BYTE, buf);
        buf.rewind();

        int pixelCount = width * height;
        int[] colors = new int[pixelCount];
        buf.asIntBuffer().get(colors);
        for (int i = 0; i < pixelCount; i++) {
            int c = colors[i];
            colors[i] = (c & 0xff00ff00) | ((c & 0x00ff0000) >> 16) | ((c & 0x000000ff) << 16);
        }

        return Bitmap.createBitmap(colors, width, height, Bitmap.Config.ARGB_8888);
    }
}