package com.baidu.cloud.videoplayer.widget;

import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.baidu.cloud.media.player.IMediaPlayer;
import com.baidu.cloud.media.player.render.record.IRecordListener;
import com.google.android.apps.muzei.render.GLTextureView;

import android.annotation.TargetApi;
import android.content.Context;
import android.graphics.SurfaceTexture;
import android.os.Build;
import android.util.AttributeSet;
import android.util.Log;
import android.view.Surface;
import android.view.TextureView;
import android.view.View;

public class VrRenderView extends GLTextureView implements IRenderView,
        TextureView.SurfaceTextureListener {
    private static final String TAG = "VrRenderView";
    private MeasureHelper mMeasureHelper;
    private Context mContext;

    public VrRenderView(Context context) {
        super(context);
        initView(context);
    }

    public VrRenderView(Context context, AttributeSet attrs) {
        super(context, attrs);
        initView(context);
    }

    public VrRenderView(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs);
        initView(context);
    }

    private void initView(Context context) {
        mContext = context;
        mMeasureHelper = new MeasureHelper(this);
        mWeakRenderView = new WeakReference<VrRenderView>(this);
        setSurfaceTextureListener(this);
    }

    @Override
    public View getView() {
        return this;
    }

    @Override
    public boolean shouldWaitForResize() {
        return false;
    }


    // --------------------
    // Layout & Measure
    // --------------------
    @Override
    public void setVideoSize(int videoWidth, int videoHeight) {
        if (videoWidth > 0 && videoHeight > 0) {
            mMeasureHelper.setVideoSize(videoWidth, videoHeight);
            requestLayout();
        }
    }

    @Override
    public void setVideoSampleAspectRatio(int videoSarNum, int videoSarDen) {
        if (videoSarNum > 0 && videoSarDen > 0) {
            mMeasureHelper.setVideoSampleAspectRatio(videoSarNum, videoSarDen);
            requestLayout();
        }
    }

    @Override
    public void setVideoRotation(int degree) {
        mMeasureHelper.setVideoRotation(degree);
        setRotation(degree);
    }

    @Override
    public void setAspectRatio(int aspectRatio) {
        mMeasureHelper.setAspectRatio(aspectRatio);
        requestLayout();
    }

    @Override
    protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        mMeasureHelper.doMeasure(widthMeasureSpec, heightMeasureSpec);
        setMeasuredDimension(mMeasureHelper.getMeasuredWidth(), mMeasureHelper.getMeasuredHeight());
    }

    private int mMediaPlayerCode = 0;

    public void setCurrentMediaPlayerCode(int currentCode) {
        mMediaPlayerCode = currentCode;
    }

    public int getCurrentMediaPlayerCode() {
        return mMediaPlayerCode;
    }

    private SurfaceTexture lastSurfaceTexture;

    public SurfaceTexture getLastSurfaceTexture() {
        return lastSurfaceTexture;
    }

    public void setLastSurfaceTexture(SurfaceTexture surfaceTexture) {
        lastSurfaceTexture = surfaceTexture;
    }

    // --------------------
    // VrRenderViewHolder
    // --------------------

    public IRenderView.ISurfaceHolder getSurfaceHolder() {
        return new VrRenderView.InternalSurfaceHolder(this);
    }

    private final class InternalSurfaceHolder implements IRenderView.ISurfaceHolder {
        private VrRenderView mTextureRenderView;

        public InternalSurfaceHolder(VrRenderView textureView) {
            mTextureRenderView = textureView;
        }

        @TargetApi(Build.VERSION_CODES.JELLY_BEAN)
        public void bindToMediaPlayer(IMediaPlayer mp) {
            if (mp == null || mTextureRenderView.getSurfaceTexture() == null) {
                // if reSetRender is invoked twice during one playing, surfaceTexture maybe null
                return;
            }
            // Tricky: 由 Object 类定义的 hashCode 方法确实会针对不同的对象返回不同的整数。
            // （这一般是通过将该对象的内部地址转换成一个整数来实现的，但是 JavaTM 编程语言不需要这种实现技巧。）
            if (mp.hashCode() != mTextureRenderView.getCurrentMediaPlayerCode()) {
                // new player
                mp.setSurface(openSurface());
            } else {
                // old player
                if (!mTextureRenderView.getLastSurfaceTexture().equals(mTextureRenderView.getSurfaceTexture())) {
                    mTextureRenderView.setSurfaceTexture(mTextureRenderView.getLastSurfaceTexture());
                }
            }
            mTextureRenderView.setCurrentMediaPlayerCode(mp.hashCode());
        }

        @Override
        public IRenderView getRenderView() {
            return mTextureRenderView;
        }

        @Override
        public Surface openSurface() {
            // TODO
            return null;
        }
    }

    // SurfaceHolder.Callback
    // -------------------------
    // -------------------------

    @Override
    public void addRenderCallback(IRenderCallback callback) {
        addRenderCallbackInternal(callback);
    }

    @Override
    public void removeRenderCallback(IRenderCallback callback) {
        removeRenderCallbackInternal(callback);
    }

    @Override
    public void release() {
        if (lastSurfaceTexture != null) {
            if (this.isAvailable()) {
                // let textureview to release surfacetexture
                setOwnSurfaceTexture(true);
            } else {
                lastSurfaceTexture.release();
                lastSurfaceTexture = null;
            }
        }
    }

    @Override
    public void destroy() {
        Log.d(TAG, "vrRenderView destroy()");
    }

    @Override
    public void startRecord(String filePath) {
        // not support
    }

    @Override
    public void setRecordListener(IRecordListener listener) {
        // not support
    }

    @Override
    public void stopRecord() {
        // not support
    }

    @Override
    public void onUpdateAudioData(byte[] bytes, int byteSize) {
        // not support
    }

    private SurfaceTexture mSurfaceTexture;
    private int mWidth;
    private int mHeight;
    private boolean mIsFormatChanged;

    private volatile boolean mOwnSurfaceTexture = false;

    private WeakReference<VrRenderView> mWeakRenderView;
    private Map<IRenderCallback, Object> mRenderCallbackMap = new ConcurrentHashMap<IRenderCallback, Object>();

    public void setOwnSurfaceTexture(boolean ownSurfaceTexture) {
        mOwnSurfaceTexture = ownSurfaceTexture;
    }

    public void addRenderCallbackInternal(IRenderCallback callback) {
        mRenderCallbackMap.put(callback, callback);

        ISurfaceHolder surfaceHolder = null;
        if (mSurfaceTexture != null) {
            if (surfaceHolder == null) {
                surfaceHolder = new VrRenderView.InternalSurfaceHolder(mWeakRenderView.get());
            }
            callback.onSurfaceCreated(surfaceHolder, mWidth, mHeight);
        }

        if (mIsFormatChanged) {
            if (surfaceHolder == null) {
                surfaceHolder = new VrRenderView.InternalSurfaceHolder(mWeakRenderView.get());
            }
            callback.onSurfaceChanged(surfaceHolder, 0, mWidth, mHeight);
        }
    }

    public void removeRenderCallbackInternal(IRenderCallback callback) {
        mRenderCallbackMap.remove(callback);
    }

    @TargetApi(16)
    @Override
    public void onSurfaceTextureAvailable(SurfaceTexture surface, int width, int height) {
        super.onSurfaceTextureAvailable(surface, width, height);
        mSurfaceTexture = surface;
        if (mWeakRenderView.get() == null) {
            // should never come in
            Log.d(TAG, "!!!!!Too bad, textureview in callback is released. function will not work normally");
        } else {
            // only set once
            if (mWeakRenderView.get().getLastSurfaceTexture() == null) {
                mWeakRenderView.get().setLastSurfaceTexture(mSurfaceTexture);
            }
        }

        mIsFormatChanged = false;
        mWidth = width;
        mHeight = height;

        Log.d(TAG, "onSurfaceTextureAvailable() " + width + "-" + height);

        ISurfaceHolder surfaceHolder = new VrRenderView.InternalSurfaceHolder(mWeakRenderView.get());
        for (IRenderCallback renderCallback : mRenderCallbackMap.keySet()) {
            renderCallback.onSurfaceCreated(surfaceHolder, 0, 0);
        }
    }

    @Override
    public void onSurfaceTextureSizeChanged(SurfaceTexture surface, int width, int height) {
        super.onSurfaceTextureSizeChanged(surface, width, height);
        mSurfaceTexture = surface;
        mIsFormatChanged = true;
        mWidth = width;
        mHeight = height;

        Log.d(TAG, "onSurfaceTextureSizeChanged() " + width + "-" + height);

        ISurfaceHolder surfaceHolder = new VrRenderView.InternalSurfaceHolder(mWeakRenderView.get());
        for (IRenderCallback renderCallback : mRenderCallbackMap.keySet()) {
            renderCallback.onSurfaceChanged(surfaceHolder, 0, width, height);
        }
    }

    @Override
    public boolean onSurfaceTextureDestroyed(SurfaceTexture surface) {
        super.onSurfaceTextureDestroyed(surface);

        mSurfaceTexture = surface;
        mIsFormatChanged = false;
        mWidth = 0;
        mHeight = 0;

        Log.d(TAG, "onSurfaceTextureDestroyed()");

        ISurfaceHolder surfaceHolder = new VrRenderView.InternalSurfaceHolder(mWeakRenderView.get());
        for (IRenderCallback renderCallback : mRenderCallbackMap.keySet()) {
            renderCallback.onSurfaceDestroyed(surfaceHolder);
        }

        return mOwnSurfaceTexture;
    }

    @Override
    public void onSurfaceTextureUpdated(SurfaceTexture surface) {
        super.onSurfaceTextureUpdated(surface);

        // Logger.e(TAG, "onSurfaceTextureUpdated()");
    }
}
